from django.urls import path
from .views import (
    view_signup_code,
    view_reset_password,
    view_invite_member,
)


urlpatterns = [
    path('signup_code.<suffix>', view_signup_code),
    path('reset_password.<suffix>', view_reset_password),
    path('invite_member.<suffix>', view_invite_member),
]



from django.utils.translation import gettext_lazy as _
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.exceptions import NotFound, PermissionDenied
from rest_framework.mixins import ListModelMixin
from ..settings import saas_settings
from ..drf.views import TenantEndpoint
from ..drf.filters import TenantIdFilter, IncludeFilter
from ..mail import SendEmailMixin
from ..serializers.member import (
    MemberSerializer,
    MemberInviteSerializer,
    MemberDetailSerializer,
    GroupPermissionSerializer,
    PermissionSerializer,
)
from ..models import Member, Group, Permission
from ..signals import member_invited


class MemberListEndpoint(SendEmailMixin, ListModelMixin, TenantEndpoint):
    email_template_id = 'invite_member'
    email_subject = _("You've Been Invited to Join %s")

    serializer_class = MemberSerializer
    filter_backends = [TenantIdFilter, IncludeFilter]
    queryset = Member.objects.all()
    resource_name = 'tenant'
    resource_http_method_actions = {
        'GET': 'read',
        'POST': 'admin',
    }
    include_select_related_fields = ['user']
    include_prefetch_related_fields = ['groups', 'permissions']

    def get_email_subject(self):
        return self.email_subject % str(self.request.tenant)

    def get(self, request: Request, *args, **kwargs):
        """List all members in the tenant."""
        return self.list(request, *args, **kwargs)

    def post(self, request: Request, *args, **kwargs):
        """Invite a member to join the tenant."""
        tenant_id = self.get_tenant_id()
        context = self.get_serializer_context()
        serializer = MemberInviteSerializer(data=request.data, context=context)
        serializer.is_valid(raise_exception=True)
        member = serializer.save(tenant_id=tenant_id, inviter=request.user)

        member_invited.send(self.__class__, member=member, request=request)
        self.send_email(
            [member.invite_email],
            inviter=request.user,
            member=member,
            tenant=request.tenant,
            invite_link=saas_settings.MEMBER_INVITE_LINK % str(member.id),
        )
        data = MemberSerializer(member).data
        return Response(data)


class MemberItemEndpoint(TenantEndpoint):
    serializer_class = MemberDetailSerializer
    queryset = Member.objects.all()
    resource_name = 'tenant'

    def get(self, request: Request, *args, **kwargs):
        """Retrieve the information of a member."""
        queryset = self.filter_queryset(self.get_queryset())
        queryset = queryset.prefetch_related('groups', 'permissions', 'groups__permissions')
        member = self.get_object_or_404(queryset, pk=kwargs['pk'])
        self.check_object_permissions(request, member)
        serializer = self.get_serializer(member)
        return Response(serializer.data)

    def delete(self, request: Request, *args, **kwargs):
        """Remove a member from the tenant."""
        queryset = self.filter_queryset(self.get_queryset())
        member = self.get_object_or_404(queryset, pk=kwargs['pk'])
        self.check_object_permissions(request, member)
        if member.is_owner:
            queryset = Member.objects.filter(tenant_id=self.get_tenant_id(), is_owner=True)
            if not queryset.count():
                raise PermissionDenied(_('The tenant should contain at lease 1 owner.'))
        member.delete()
        return Response(status=204)


class _MemberEndpoint(TenantEndpoint):
    resource_http_method_actions = {
        'GET': 'read',
        'POST': 'admin',
        'DELETE': 'admin',
    }

    def get_member(self, member_id) -> Member:
        try:
            member = Member.objects.get_from_cache_by_pk(member_id)
            if member.tenant_id != self.get_tenant_id():
                raise NotFound()
        except Member.DoesNotExist:
            raise NotFound()
        self.check_object_permissions(self.request, member)
        return member

    def list(self, request: Request, *args, **kwargs):
        member = self.get_member(kwargs['member_id'])
        queryset = self.get_queryset().filter(member=member)
        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    def replace(self, request: Request, field_name: str, **kwargs):
        member = self.get_member(kwargs['member_id'])
        queryset = self.filter_queryset(self.get_queryset())
        items = queryset.filter(pk__in=request.data).all()
        getattr(member, field_name).set(items)
        return Response(status=204)

    def destroy(self, request: Request, field_name: str, **kwargs):
        obj = self.get_object_or_404(self.get_queryset(), pk=kwargs['pk'])
        member = self.get_member(kwargs['member_id'])
        if obj.member_id != member.pk:
            raise NotFound()
        getattr(member, field_name).remove(obj)
        return Response(status=204)


class MemberGroupsEndpoint(_MemberEndpoint):
    serializer_class = GroupPermissionSerializer
    queryset = Group.objects.all()

    def get(self, request: Request, *args, **kwargs):
        """List all groups of the selected member."""
        return self.list(request, *args, **kwargs)

    def post(self, request: Request, *args, **kwargs):
        """Reset groups of the selected member."""
        return self.replace(request, field_name='groups', **kwargs)


class MemberPermissionsEndpoint(_MemberEndpoint):
    serializer_class = PermissionSerializer
    queryset = Permission.objects.all()

    def get(self, request: Request, *args, **kwargs):
        """List all permissions of the selected member."""
        return self.list(request, *args, **kwargs)

    def post(self, request: Request, *args, **kwargs):
        """Reset permissions of the selected member."""
        return self.replace(request, field_name='permissions', **kwargs)


class MemberGroupItemEndpoint(_MemberEndpoint):
    serializer_class = GroupPermissionSerializer
    queryset = Group.objects.all()

    def delete(self, request: Request, *args, **kwargs):
        """Remove a group of the selected member."""
        return self.destroy(request, field_name='groups', **kwargs)


class MemberPermissionItemEndpoint(_MemberEndpoint):
    serializer_class = PermissionSerializer
    queryset = Permission.objects.all()

    def delete(self, request: Request, *args, **kwargs):
        """Remove a permission of the selected member."""
        return self.destroy(request, field_name='permissions', **kwargs)



from django.urls import path
from ..endpoints.user import (
    UserEndpoint,
    UserPasswordEndpoint,
    UserEmailListEndpoint,
    UserTenantsEndpoint,
)

urlpatterns = [
    path('', UserEndpoint.as_view()),
    path('password/', UserPasswordEndpoint.as_view()),
    path('emails/', UserEmailListEndpoint.as_view()),
    path('tenants/', UserTenantsEndpoint.as_view()),
]



from django.apps import AppConfig
from django.conf import settings
from django.db.models.signals import post_migrate
from .permissions import create_permissions

if not hasattr(settings, 'SAAS_TENANT_MODEL'):
    setattr(settings, 'SAAS_TENANT_MODEL', 'saas_base.Tenant')


class CoreConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'saas_base'
    verbose_name = 'SaaS'

    def ready(self):
        post_migrate.connect(create_permissions)



from typing import Type, Optional
from django.apps import apps
from django.db import models
from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from .member import Member
from ..db import CachedManager
from ..settings import saas_settings


class TenantManager(CachedManager):
    natural_key = ['slug']

    def create(self, slug: str, owner: Optional[AbstractUser] = None, **kwargs):
        kwargs.setdefault('region', saas_settings.DEFAULT_REGION)
        tenant = super().create(slug=slug, owner=owner, **kwargs)
        # initial with owner membership
        if owner:
            name = owner.get_full_name() or owner.get_username()
            Member.objects.update_or_create(
                user=owner,
                tenant=tenant,
                defaults={
                    'name': name,
                    'is_owner': True,
                    'status': Member.InviteStatus.ACTIVE,
                },
            )
        return tenant

    def get_by_slug(self, slug: str):
        return self.get_from_cache_by_natural_key(slug)


class AbstractTenant(models.Model):
    slug = models.SlugField(
        unique=True,
        help_text='Identity of the tenant, e.g. <slug>.example.com',
    )
    region = models.CharField(max_length=24, blank=True, default='')
    environment = models.CharField(max_length=48, blank=True, default='')
    expires_at = models.DateTimeField(null=True, blank=True)
    created_at = models.DateTimeField(default=timezone.now, db_index=True)
    updated_at = models.DateTimeField(auto_now=True)
    owner = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, on_delete=models.SET_NULL)

    objects = TenantManager()

    class Meta:
        verbose_name = _('tenant')
        verbose_name_plural = _('tenants')
        abstract = True

    def __str__(self):
        return self.slug

    def natural_key(self):
        return (self.slug,)


class Tenant(AbstractTenant):
    objects = TenantManager['Tenant']()

    class Meta(AbstractTenant.Meta):
        swappable = 'SAAS_TENANT_MODEL'
        db_table = 'saas_tenant'
        ordering = ['created_at']


def get_tenant_model() -> Type[Tenant]:
    return apps.get_model(settings.SAAS_TENANT_MODEL)



from saas_base.models import UserEmail
from tests.client import FixturesTestCase


class TestEmailAPI(FixturesTestCase):
    user_id = FixturesTestCase.EMPTY_USER_ID

    def test_list_emails(self):
        self.force_login()

        url = '/m/user/emails/'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json(), [])

        for i in range(5):
            UserEmail.objects.create(user=self.user, email=f'demo-{self.user_id}-{i}@example.com')

        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(resp.json()), 5)



import typing as t
import requests
from .base import BaseMailProvider


class MailgunProvider(BaseMailProvider):
    name = 'mailgun'

    def get_api_url(self) -> str:
        domain = self.options['domain']
        return f'https://api.mailgun.net/v3/{domain}/messages'

    def get_api_auth(self):
        return 'api', self.options['api_key']

    def send_mail(
        self,
        subject: str,
        recipients: t.List[str],
        text_message: str,
        html_message: t.Optional[str] = None,
        from_email: t.Optional[str] = None,
        headers: t.Optional[t.Dict[str, str]] = None,
        reply_to: t.Optional[str] = None,
        fail_silently: bool = False,
    ):
        if from_email is None:
            from_email = self.default_from_email
        url = self.get_api_url()
        auth = self.get_api_auth()
        data = {
            'from': from_email,
            'to': recipients,
            'subject': subject,
            'text': text_message,
            'html': html_message,
        }
        if reply_to:
            data['h:Reply-To'] = reply_to
        for k in headers:
            data[f'h:{k}'] = headers[k]
        return requests.post(url, auth=auth, data=data)



from django.urls import path, include
from ..endpoints.tenant import TenantListEndpoint

urlpatterns = [
    path('tenants/', TenantListEndpoint.as_view()),
    path('user/', include('saas_base.api_urls.user')),
    path('tenant/', include('saas_base.api_urls.tenant')),
    path('members/', include('saas_base.api_urls.members')),
    path('session/', include('saas_base.api_urls.session')),
]



import random
import string
import typing as t
from email.utils import formataddr
from django.db import transaction
from django.core.cache import cache
from django.utils.translation import gettext_lazy as _
from django.contrib.auth import password_validation, get_user_model
from django.contrib.auth.validators import UnicodeUsernameValidator
from django.contrib.auth.models import User
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from ..models import UserEmail

SIGNUP_CODE = 'saas:signup_code'

ERRORS = {
    'email': _('This email address is already associated with an existing account.'),
    'username': _('This username is already associated with an existing account.'),
    'code': _('Code does not match or expired.'),
}


class EmailCode:
    def __init__(self, email: str, code: str, user=None):
        self.email = email
        self.code = code
        self.user = user

    def recipient(self):
        if self.user:
            return formataddr((self.user.username, self.email))
        return self.email


class SignupRequestCodeSerializer(serializers.Serializer):
    email = serializers.EmailField(required=True)

    def validate_email(self, email: str):
        try:
            UserEmail.objects.get(email=email)
            raise ValidationError(ERRORS['email'])
        except UserEmail.DoesNotExist:
            return email

    def create(self, validated_data) -> EmailCode:
        email = validated_data['email']
        code = ''.join(random.sample(string.ascii_uppercase, 6))
        cache_key = f'{SIGNUP_CODE}:{email}:{code}'
        cache.set(cache_key, 1, timeout=300)
        return EmailCode(email, code)


class SignupCreateUserSerializer(serializers.Serializer):
    username = serializers.CharField(required=True, validators=[UnicodeUsernameValidator()])
    email = serializers.EmailField(required=True)
    password = serializers.CharField(required=True)

    def validate_email(self, email: str):
        try:
            UserEmail.objects.get(email=email)
            raise ValidationError(ERRORS['email'])
        except UserEmail.DoesNotExist:
            return email

    def validate_username(self, username: str):
        cls: t.Type[User] = get_user_model()
        try:
            cls.objects.get(username=username)
            raise ValidationError(ERRORS['username'])
        except cls.DoesNotExist:
            return username

    def validate_password(self, raw_password: str):
        user = User(
            username=self.initial_data['username'],
            email=self.initial_data['email'],
        )
        password_validation.validate_password(raw_password, user)
        return raw_password

    def create(self, validated_data) -> EmailCode:
        username = validated_data['username']
        email = validated_data['email']
        password = validated_data['password']
        cls: t.Type[User] = get_user_model()
        with transaction.atomic():
            user = cls.objects.create_user(
                username=username,
                email=email,
                password=password,
                is_active=False,
            )
            UserEmail.objects.create(user=user, email=email, primary=True, verified=False)

        code = ''.join(random.sample(string.ascii_uppercase, 6))
        cache_key = f'{SIGNUP_CODE}:{email}:{code}'
        cache.set(cache_key, 1, timeout=300)
        return EmailCode(email, code, user)


class SignupConfirmCodeSerializer(serializers.Serializer):
    email = serializers.EmailField(required=True)
    code = serializers.CharField(required=True, max_length=6)

    def validate_code(self, code: str):
        email = self.initial_data['email']
        code = code.upper()
        cache_key = f'{SIGNUP_CODE}:{email}:{code}'
        has_code: str = cache.get(cache_key)
        if not has_code:
            raise ValidationError(ERRORS['code'])

        cache.delete(cache_key)
        try:
            return UserEmail.objects.select_related('user').get(email=email)
        except UserEmail.DoesNotExist:
            raise ValidationError(ERRORS['code'])

    def create(self, validated_data) -> User:
        user_email = validated_data['code']
        with transaction.atomic():
            user_email.verified = True
            user = user_email.user
            user.is_active = True
            user_email.save()
            user.save()
        return user


class SignupConfirmPasswordSerializer(serializers.Serializer):
    username = serializers.CharField(required=True, validators=[UnicodeUsernameValidator()])
    email = serializers.EmailField(required=True)
    code = serializers.CharField(required=True, max_length=6)
    password = serializers.CharField(required=True)

    def validate_username(self, username: str):
        cls: t.Type[User] = get_user_model()
        try:
            cls.objects.get(username=username)
            raise ValidationError(ERRORS['username'])
        except cls.DoesNotExist:
            return username

    def validate_password(self, raw_password: str):
        password_validation.validate_password(raw_password)
        return raw_password

    def validate_code(self, code: str):
        email = self.initial_data['email']
        code = code.upper()
        cache_key = f'{SIGNUP_CODE}:{email}:{code}'
        has_code: str = cache.get(cache_key)
        if not has_code:
            raise ValidationError(ERRORS['code'])
        return code

    def create(self, validated_data) -> User:
        username = validated_data['username']
        email = validated_data['email']
        password = validated_data['password']
        cls: t.Type[User] = get_user_model()
        with transaction.atomic():
            user = cls.objects.create_user(
                username=username,
                email=email,
                password=password,
                is_active=True,
            )
            UserEmail.objects.create(user=user, email=email, primary=True, verified=True)

        code = validated_data['code']
        cache_key = f'{SIGNUP_CODE}:{email}:{code}'
        cache.delete(cache_key)
        return user



from drf_spectacular.extensions import OpenApiSerializerFieldExtension


class ChoiceFieldFix(OpenApiSerializerFieldExtension):
    target_class = 'saas_base.drf.serializers.ChoiceField'

    def map_serializer_field(self, auto_schema, direction):
        choices = list(self.target.choices.values())
        return {'type': 'string', 'enum': choices}



from typing import List

from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from ..db import CachedManager
from ..settings import saas_settings


class PermissionManager(CachedManager['Permission']):
    natural_key = ['name']

    def get_by_name(self, name: str):
        return self.get_from_cache_by_pk(name)

    def initialize_names(self, resources: List[str]):
        objs = []
        for resource in resources:
            for action in self.model.actions:
                name = saas_settings.PERMISSION_NAME_FORMATTER.format(resource=resource, action=action)
                objs.append(self.model(name=name, description=''))

        return self.bulk_create(objs, ignore_conflicts=True)


class Permission(models.Model):
    actions = ['read', 'write', 'admin']

    name = models.CharField(max_length=100, primary_key=True, editable=False)
    description = models.CharField(max_length=480, blank=True)
    # is this permission designed for internal use only?
    internal = models.BooleanField(default=False, db_index=True)
    created_at = models.DateTimeField(default=timezone.now)

    objects = PermissionManager()

    class Meta:
        verbose_name = _('permission')
        verbose_name_plural = _('permissions')
        db_table = 'saas_permission'
        ordering = ['name']

    def __str__(self):
        return self.name

    def natural_key(self):
        return (self.name,)



import random
import string
from django.utils.translation import gettext as _
from django.core.cache import cache
from django.contrib.auth import password_validation, authenticate
from django.contrib.auth.models import AbstractUser
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from ..models import UserEmail

CACHE_PREFIX = 'saas:password_code'


class PasswordLoginSerializer(serializers.Serializer):
    username = serializers.CharField(required=True)
    password = serializers.CharField(required=True)

    @staticmethod
    def invalid_errors():
        errors = {'password': [_('Invalid username or password.')]}
        return ValidationError(errors)

    def create(self, validated_data):
        request = self.context['request']
        user = authenticate(request=request, **validated_data)
        if not user:
            raise self.invalid_errors()
        return user

    def update(self, instance, validated_data):
        raise RuntimeError('This method is not allowed.')


class PasswordForgetSerializer(serializers.Serializer):
    email = serializers.EmailField(required=True)

    @staticmethod
    def save_password_code(obj: UserEmail) -> str:
        code = ''.join(random.sample(string.ascii_uppercase, 6))
        key = f'{CACHE_PREFIX}:{obj.email}:{code}'
        cache.set(key, obj.user_id, timeout=300)
        return code

    def create(self, validated_data) -> UserEmail:
        email = validated_data['email']
        try:
            obj = UserEmail.objects.get(email=email)
        except UserEmail.DoesNotExist:
            raise ValidationError({'email': [_('Invalid email address.')]})
        return obj


class PasswordResetSerializer(PasswordForgetSerializer):
    code = serializers.CharField(required=True, max_length=6)
    password = serializers.CharField(required=True)

    def validate_password(self, raw_password):
        password_validation.validate_password(raw_password)
        return raw_password

    def create(self, validated_data):
        obj = super().create(validated_data)

        code = validated_data['code']
        key = f'{CACHE_PREFIX}:{obj.email}:{code}'
        user_id = cache.get(key)
        if not user_id or obj.user_id != user_id:
            raise ValidationError({'code': [_('Code does not match or expired.')]})

        raw_password = validated_data['password']
        user: AbstractUser = obj.user
        user.set_password(raw_password)
        user.save()

        cache.delete(key)
        return obj



import os
import json
from saas_base.test import SaasTestCase
from requests_mock import Mocker

ROOT = os.path.dirname(__file__)


class FixturesTestCase(SaasTestCase):
    @staticmethod
    def load_fixture(name: str):
        filename = os.path.join(ROOT, 'fixtures', name)
        with open(filename) as f:
            data = json.load(f)
        return data

    @classmethod
    def mock_requests(cls, *names: str):
        m = Mocker()
        for name in names:
            data = cls.load_fixture(name)
            m.register_uri(**data)
        return m



from django.utils.translation import gettext as _
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser
from django.contrib.auth import password_validation
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from ..models import UserEmail


class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = get_user_model()
        exclude = ['password', 'groups', 'user_permissions']


class UserEmailSerializer(serializers.ModelSerializer):
    class Meta:
        model = UserEmail
        exclude = ['user']


class UserPasswordSerializer(serializers.Serializer):
    password = serializers.CharField(required=True)
    confirm_password = serializers.CharField(required=True)

    def validate_password(self, raw_password: str):
        if self.initial_data['confirm_password'] != raw_password:
            raise ValidationError(_('Password does not match.'))
        password_validation.validate_password(raw_password)
        return raw_password

    def update(self, user: AbstractUser, validated_data):
        raw_password = validated_data['password']
        user.set_password(raw_password)
        user.save()
        return user



from rest_framework.permissions import BasePermission, SAFE_METHODS
from rest_framework.request import Request
from ..models import get_tenant_model, Member
from ..settings import saas_settings

__all__ = [
    'IsTenantOwner',
    'IsTenantOwnerOrReadOnly',
    'HasResourcePermission',
    'HasResourceScope',
]

TenantModel = get_tenant_model()

http_method_actions = {
    'GET': 'read',
    'HEAD': 'read',
    'POST': 'write',
    'PUT': 'write',
    'PATCH': 'write',
    'DELETE': 'admin',
}


class IsTenantOwner(BasePermission):
    """The authenticated user is the tenant owner."""

    def has_permission(self, request: Request, view):
        tenant_id = getattr(request, 'tenant_id', None)
        if not tenant_id:
            return False
        try:
            tenant = TenantModel.objects.get_from_cache_by_pk(tenant_id)
            return request.user.pk == tenant.owner_id
        except TenantModel.DoesNotExist:
            return False


class IsTenantOwnerOrReadOnly(IsTenantOwner):
    """The authenticated user is the tenant owner, or is a read-only request."""

    def has_permission(self, request, view):
        if request.method in SAFE_METHODS:
            return True
        return super().has_permission(request, view)


class HasResourcePermission(BasePermission):
    """The authenticated user is a member of the tenant, and the user
    has the given resource permission.
    """

    @staticmethod
    def get_resource_permissions(view, method):
        resource = getattr(view, 'resource_name', None)
        if not resource:
            return

        action = getattr(view, 'resource_action', None)
        if not action:
            method_actions = getattr(view, 'resource_http_method_actions', http_method_actions)
            action = method_actions.get(method)

        permission = saas_settings.PERMISSION_NAME_FORMATTER.format(
            resource=resource,
            action=action,
        )
        return [permission]

    @staticmethod
    def get_tenant(request: Request):
        tenant_id = getattr(request, 'tenant_id', None)
        if not tenant_id:
            return None
        try:
            return TenantModel.objects.get_from_cache_by_pk(tenant_id)
        except TenantModel.DoesNotExist:
            return None

    def has_permission(self, request: Request, view):
        resource_permissions = self.get_resource_permissions(view, request.method)
        if not resource_permissions:
            return True

        # HasResourcePermission should apply to TenantEndpoint
        tenant = self.get_tenant(request)
        if not tenant:
            return False

        # Tenant owner has all permissions
        if tenant.owner_id == request.user.pk:
            return True

        try:
            member = Member.objects.get_by_natural_key(tenant.pk, request.user.pk)
            if not member.is_active:
                return False
            # owner has all permissions
            if member.is_owner:
                return True
        except Member.DoesNotExist:
            return False

        perms = member.get_all_permissions()
        if not perms:
            return False

        for name in resource_permissions:
            if name in perms:
                return True
        return False


class HasResourceScope(BasePermission):
    """The request token contains the given resource scopes."""

    @staticmethod
    def get_resource_scopes(view, method):
        if hasattr(view, 'get_resource_scopes'):
            resource_scopes = view.get_resource_scopes(method)
        elif hasattr(view, 'resource_scopes'):
            resource_scopes = view.resource_scopes
        else:
            resource_scopes = None
        return resource_scopes

    def has_permission(self, request: Request, view):
        resource_scopes = self.get_resource_scopes(view, request.method)
        if not resource_scopes:
            return True

        if request.auth is None:
            return True

        scope = getattr(request.auth, 'scope', '')
        if scope == '__all__':
            return True

        token_scopes = set(scope.split())
        for rs in resource_scopes:
            if set(rs.split()).issubset(token_scopes):
                return True
        return False



from .permission import Permission
from .tenant import AbstractTenant, Tenant, get_tenant_model
from .group import Group
from .member import Member
from .user_email import UserEmail

__all__ = [
    'Permission',
    'AbstractTenant',
    'Tenant',
    'get_tenant_model',
    'Group',
    'Member',
    'UserEmail',
]



import typing as t
import hashlib
import logging
import uuid

from django.db.models import Model, Manager, UUIDField
from django.db.models.signals import (
    class_prepared,
    post_save,
    post_delete,
)
from .cache import cache

logger = logging.getLogger(__name__)

M = t.TypeVar('M', bound=Model)


class CachedManager(Manager, t.Generic[M]):
    db_cache = cache
    natural_key: t.List[str] = []
    cache_ttl: int = 300
    cache_version = 1
    query_select_related: t.List[str] = []

    def contribute_to_class(self, model: t.Type[M], name: str):
        super().contribute_to_class(model, name)
        class_prepared.connect(self.__manage_cache, sender=model)

    def get_many_from_cache(self, pk_set: t.List[t.Any]) -> t.Dict[t.Any, M]:
        key_map = {self.__get_lookup_cache_key(pk=pk): pk for pk in pk_set}
        results = self.db_cache.get_many(key_map.keys(), version=self.cache_version)

        found = {}
        missed = []
        for key in key_map:
            if key not in results:
                missed.append(key_map[key])
            else:
                found[key_map[key]] = results[key]

        if missed:
            queryset = self.filter(pk__in=missed).select_related(*self.query_select_related).all()
            for instance in queryset:
                self.__post_save(instance)
                found[instance.pk] = instance
        return found

    def get_from_cache_by_pk(self, pk: t.Any) -> M:
        key = self.__get_lookup_cache_key(pk=pk)
        instance = self.__get_from_cache_or_raise(key)
        if instance:
            instance._state.db = self.db
            return instance

        try:
            if self.query_select_related:
                instance = self.select_related(*self.query_select_related).get(pk=pk)
            else:
                instance = self.get(pk=pk)
        except self.model.DoesNotExist:
            self.__set_not_exist_cache(key)
            raise self.model.DoesNotExist

        self.__post_save(instance)
        return instance

    def get_from_cache_by_natural_key(self, *args) -> M:
        kwargs = dict(zip(self.natural_key, args))
        key = self.__get_lookup_cache_key(**kwargs)
        pk_val = self.__get_from_cache_or_raise(key)
        if pk_val:
            return self.get_from_cache_by_pk(pk_val)

        try:
            instance = self.get_from_db_by_natural_key(**kwargs)
        except self.model.DoesNotExist:
            self.__set_not_exist_cache(key)
            raise self.model.DoesNotExist

        self.__post_save(instance)
        return instance

    def get_from_db_by_natural_key(self, **kwargs) -> M:
        if self.query_select_related:
            instance = self.select_related(*self.query_select_related).get(**kwargs)
        else:
            instance = self.get(**kwargs)
        return instance

    def purge(self, pk_value):
        key = self.__get_lookup_cache_key(pk=pk_value)
        self.db_cache.delete(key, version=self.cache_version)

    def __manage_cache(self, sender, **kwargs):
        post_save.connect(self.__post_save, sender=sender, weak=False)
        post_delete.connect(self.__post_delete, sender=sender, weak=False)

    def __get_from_cache_or_raise(self, key: str):
        value = self.db_cache.get(key, version=self.cache_version)
        if value == '__none__':
            raise self.model.DoesNotExist
        return value

    def __set_not_exist_cache(self, key: str):
        self.db_cache.set(
            key=key,
            value='__none__',
            timeout=self.cache_ttl,
            version=self.cache_version,
        )

    def __get_lookup_cache_key(self, **kwargs) -> str:
        key = make_key(self.model, kwargs)
        return f'db:{self.model._meta.db_table}:{key}'

    def __get_natural_cache_key(self, instance: M) -> str:
        natural_fields = {key: value_for_field(instance, key) for key in self.natural_key}
        return self.__get_lookup_cache_key(**natural_fields)

    def __post_save(self, instance, **kwargs):
        if not kwargs.get('created'):
            _natural_cache_key = self.db_cache.get(
                self.__get_lookup_cache_key(__track=instance.pk),
                version=self.cache_version,
            )
        else:
            _natural_cache_key = None

        natural_cache_key = self.__get_natural_cache_key(instance)
        if _natural_cache_key and _natural_cache_key != natural_cache_key:
            self.db_cache.delete(_natural_cache_key, version=self.cache_version)

        to_save = {
            natural_cache_key: instance.pk,
            self.__get_lookup_cache_key(__track=instance.pk): natural_cache_key,
        }
        self.db_cache.set_many(to_save, timeout=self.cache_ttl, version=self.cache_version)

        # Ensure we don't serialize the database into the cache
        db = instance._state.db
        instance._state.db = None

        # store actual object
        try:
            self.db_cache.set(
                key=self.__get_lookup_cache_key(pk=instance.pk),
                value=instance,
                timeout=self.cache_ttl,
                version=self.cache_version,
            )
        except Exception as e:
            logger.error(e, exc_info=True)

        # recover database on instance
        instance._state.db = db

    def __post_delete(self, instance, **kwargs):
        to_delete = [
            self.__get_lookup_cache_key(pk=instance.pk),
            self.__get_natural_cache_key(instance),
            self.__get_lookup_cache_key(__track=instance.pk),
        ]
        self.db_cache.delete_many(to_delete, version=self.cache_version)


def make_key(cls: Model, kwargs: t.Mapping[str, t.Union[Model, int, str]]) -> str:
    fields = []
    for k, v in sorted(kwargs.items()):
        if k == 'pk':
            # convert pk to its real name
            k = str(cls._meta.pk.name)
            if isinstance(v, str) and isinstance(cls._meta.pk, UUIDField):
                v = v.replace('-', '')
        if isinstance(v, Model):
            v = v.pk
        if isinstance(v, uuid.UUID):
            v = v.hex
        fields.append(f'{k}={v}')
    return hashlib.md5('&'.join(fields).encode('utf-8')).hexdigest()


def value_for_field(instance: M, key: str) -> t.Any:
    field = instance._meta.get_field(key)
    return getattr(instance, field.attname)



from django.apps import AppConfig


class RestFrameworkConfig(AppConfig):
    name = 'saas_base.drf'

    def ready(self):
        from django.apps import apps
        from django.utils.module_loading import autodiscover_modules

        try:
            apps.get_app_config('drf_spectacular')
            autodiscover_modules('spectacular_schema')
        except LookupError:
            pass



from django.urls import path
from ..endpoints.tenant import (
    SelectedTenantEndpoint,
    CurrentMemberEndpoint,
)

urlpatterns = [
    path('', SelectedTenantEndpoint.as_view()),
    path('member/', CurrentMemberEndpoint.as_view()),
]



import typing as t
import css_inline
from django.template import loader
from abc import ABCMeta, abstractmethod


class BaseMailProvider(metaclass=ABCMeta):
    name: str = 'base'

    def __init__(self, **options):
        self.default_from_email = options.pop('default_from_email', None)
        self.options = options

    @abstractmethod
    def send_mail(
        self,
        subject: str,
        recipients: t.List[str],
        text_message: str,
        html_message: t.Optional[str] = None,
        from_email: t.Optional[str] = None,
        headers: t.Optional[t.Dict[str, str]] = None,
        reply_to: t.Optional[str] = None,
        fail_silently: bool = False,
    ):
        pass

    def send_context_mail(
        self,
        subject: str,
        recipients: t.List[str],
        template_id: str,
        context: t.Dict[str, t.Any],
        from_email: t.Optional[str] = None,
        headers: t.Optional[t.Dict[str, str]] = None,
        reply_to: t.Optional[str] = None,
        fail_silently: bool = False,
    ):
        if from_email is None:
            from_email = self.default_from_email
        text_message, html_message = render_mail_messages(template_id, context)
        return self.send_mail(
            subject=subject,
            recipients=recipients,
            text_message=text_message,
            html_message=html_message,
            from_email=from_email,
            headers=headers,
            reply_to=reply_to,
            fail_silently=fail_silently,
        )


def render_mail_messages(
    template_id: str, context: t.Dict[str, t.Any], using: t.Optional[str] = None
) -> t.Tuple[str, str]:
    text: str = loader.render_to_string(f'saas_emails/{template_id}.text', context, using=using)
    html: str = loader.render_to_string(f'saas_emails/{template_id}.html', context, using=using)
    return text, css_inline.inline(html)



from typing import List, Any
from drf_spectacular.openapi import AutoSchema as _AutoSchema


class AutoSchema(_AutoSchema):
    def get_filter_backends(self) -> List[Any]:
        return getattr(self.view, 'filter_backends', [])

    def get_description(self) -> str:
        description = super().get_description()

        permissions = []
        scopes = []
        for perm in self.view.permission_classes:
            if hasattr(perm, 'get_resource_permissions'):
                permissions = perm.get_resource_permissions(self.view, self.method)
            if hasattr(perm, 'get_resource_scopes'):
                scopes = perm.get_resource_scopes(self.view, self.method)

        if permissions:
            permissions_string = ' '.join([f'`{p}`' for p in permissions])
            description = f'**Permissions**: {permissions_string}\n\n{description}'
        if scopes:
            scopes_string = ' '.join([f'`{p}`' for p in scopes])
            description = f'**Scopes**: {scopes_string}\n\n{description}'
        return description



import typing as t
from django.apps import apps as global_apps
from django.db import DEFAULT_DB_ALIAS


DEFAULT_PERMISSIONS = [
    ('tenant.read', 'Read permission for tenants'),
    ('tenant.write', 'Write permission for tenants'),
    ('tenant.admin', 'Admin permission for tenants'),
]


def create_permissions_receiver(permissions: t.List[t.Tuple[str, str]]):
    def _create_permissions(app_config, verbosity=2, using=DEFAULT_DB_ALIAS, apps=global_apps, **kwargs):
        if not app_config.models_module:
            return

        if app_config.label != 'saas_base':
            return

        try:
            Permission = apps.get_model(app_config.label, 'Permission')
        except LookupError:
            return

        existed_perms = set(Permission.objects.values_list('name', flat=True).all())

        to_add_perms = []
        for name, description in permissions:
            if name not in existed_perms:
                to_add_perms.append(Permission(name=name, description=description))

        if to_add_perms:
            Permission.objects.using(using).bulk_create(to_add_perms, ignore_conflicts=True)
        if verbosity >= 2:
            for perm in to_add_perms:
                print(f"Adding saas_base.Permission '{perm.name}'")

    return _create_permissions


create_permissions = create_permissions_receiver(DEFAULT_PERMISSIONS)



from tests.client import FixturesTestCase
from saas_base.models import Member, Group


class TestMembersAPI(FixturesTestCase):
    user_id = FixturesTestCase.GUEST_USER_ID

    def create_demo_members(self, count: int = 10):
        users = []
        for i in range(count):
            user = Member.objects.create(
                name=f'member-{i}',
                tenant=self.tenant,
            )
            users.append(user)
        return users

    def test_list_users_via_owner(self):
        self.force_login(self.OWNER_USER_ID)
        resp = self.client.get('/m/members/')
        self.assertEqual(resp.status_code, 200)

    def test_list_users_via_guest_user(self):
        self.force_login()
        resp = self.client.get('/m/members/')
        self.assertEqual(resp.status_code, 403)

        self.add_user_perms('tenant.read')
        resp = self.client.get('/m/members/')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        member = data['results'][0]
        self.assertNotIn('user', member)

    def test_list_include_user(self):
        self.add_user_perms('tenant.read')
        self.force_login()
        resp = self.client.get('/m/members/?include=user')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        member = data['results'][0]
        self.assertIn('user', member)

    def test_list_include_permissions(self):
        self.add_user_perms('tenant.read')
        self.force_login()
        resp = self.client.get('/m/members/?include=permissions')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        member = data['results'][0]
        self.assertIn('permissions', member)

    def test_invite_member_signup(self):
        self.add_user_perms('tenant.admin')
        self.force_login()
        data = {'email': 'signup@example.com'}
        resp = self.client.post('/m/members/', data=data)
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.json(), {'invite_email': ['This field is required.']})

        # invite non-exist user
        data = {'invite_email': 'signup@example.com'}
        resp = self.client.post('/m/members/', data=data)
        self.assertEqual(resp.status_code, 200)
        member = resp.json()
        self.assertEqual(member['status'], 'request')

        # invite existing user
        user = self.get_user(self.STAFF_USER_ID)
        data = {'invite_email': user.email}
        resp = self.client.post('/m/members/', data=data)
        self.assertEqual(resp.status_code, 200)
        member = resp.json()
        self.assertEqual(member['status'], 'waiting')

    def test_view_member_item(self):
        self.add_user_perms('tenant.read')
        self.force_login()
        member = Member.objects.filter(tenant=self.tenant, user_id=self.user_id).first()
        resp = self.client.get(f'/m/members/{member.id}/')
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertIn('groups', data)
        self.assertIn('permissions', data)

    def test_remove_member_item(self):
        self.add_user_perms('tenant.admin')
        self.force_login()
        member = Member.objects.filter(tenant=self.tenant, user_id=self.user_id).first()
        resp = self.client.delete(f'/m/members/{member.id}/')
        self.assertEqual(resp.status_code, 204)

    def test_list_member_permissions(self):
        self.add_user_perms('tenant.read')
        self.force_login()
        member = Member.objects.filter(tenant=self.tenant, user_id=self.user_id).first()
        resp = self.client.get(f'/m/members/{member.id}/permissions/')
        self.assertEqual(resp.status_code, 200)

    def test_list_member_groups(self):
        self.add_user_perms('tenant.read')
        self.force_login()
        member = Member.objects.filter(tenant=self.tenant, user_id=self.user_id).first()
        resp = self.client.get(f'/m/members/{member.id}/groups/')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json(), [])

        groups = Group.objects.filter(tenant=self.tenant).all()
        member.groups.set(groups)
        resp = self.client.get(f'/m/members/{member.id}/groups/')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(resp.json()), 1)
        permissions = resp.json()[0]['permissions']
        self.assertEqual(len(permissions), 3)



from django.utils.translation import gettext_lazy as _
from django.conf import settings
from django.contrib.auth import login, logout
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.throttling import AnonRateThrottle
from ..models import Member
from ..drf.views import Endpoint
from ..settings import saas_settings
from ..security import check_security_rules
from ..serializers.auth import (
    EmailCode,
    SignupRequestCodeSerializer,
    SignupCreateUserSerializer,
    SignupConfirmCodeSerializer,
    SignupConfirmPasswordSerializer,
)
from ..serializers.password import PasswordLoginSerializer
from ..signals import after_signup_user, after_login_user
from ..mail import SendEmailMixin


class SignupRequestEndpoint(SendEmailMixin, Endpoint):
    email_template_id = 'signup_code'
    email_subject = _('Signup Request')
    authentication_classes = []
    permission_classes = []
    throttle_classes = [AnonRateThrottle]

    def get_serializer_class(self):
        if saas_settings.SIGNUP_REQUEST_CREATE_USER:
            return SignupCreateUserSerializer
        return SignupRequestCodeSerializer

    def post(self, request: Request):
        """Send a sign-up code to user's email address."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj: EmailCode = serializer.save()
        # check bad request rules
        check_security_rules(saas_settings.SIGNUP_SECURITY_RULES, request)

        self.send_email([obj.recipient()], code=obj.code)
        return Response(status=204)


class SignupConfirmEndpoint(Endpoint):
    authentication_classes = []
    permission_classes = []
    throttle_classes = [AnonRateThrottle]

    def get_serializer_class(self):
        if saas_settings.SIGNUP_REQUEST_CREATE_USER:
            return SignupConfirmCodeSerializer
        return SignupConfirmPasswordSerializer

    def post(self, request: Request):
        """Register a new user and login."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = serializer.save()

        # update related membership
        Member.objects.filter(invite_email=user.email).update(user=user, status=Member.InviteStatus.WAITING)
        after_signup_user.send(
            self.__class__,
            user=user,
            request=request,
            strategy='password',
        )
        return Response({'next': settings.LOGIN_REDIRECT_URL})


class PasswordLogInEndpoint(Endpoint):
    authentication_classes = []
    permission_classes = []
    throttle_classes = [AnonRateThrottle]
    serializer_class = PasswordLoginSerializer

    def post(self, request: Request):
        """Login a user with the given username and password."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = serializer.save()
        login(request._request, user)

        after_login_user.send(
            self.__class__,
            user=user,
            request=request,
            strategy='password',
        )
        return Response({'next': settings.LOGIN_REDIRECT_URL})


class LogoutEndpoint(Endpoint):
    authentication_classes = []
    permission_classes = []

    def post(self, request: Request):
        """Clear the user session and log the user out."""
        logout(request._request)
        return Response({'next': settings.LOGIN_URL})



from email.utils import formataddr
from django.utils.translation import gettext_lazy as _
from django.conf import settings
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.throttling import AnonRateThrottle
from ..drf.views import Endpoint
from ..settings import saas_settings
from ..models import UserEmail
from ..mail import SendEmailMixin
from ..security import check_security_rules
from ..serializers.password import (
    PasswordForgetSerializer,
    PasswordResetSerializer,
)


class PasswordForgotEndpoint(SendEmailMixin, Endpoint):
    email_template_id = 'reset_password'
    email_subject = _('Password Reset Request')

    authentication_classes = []
    permission_classes = []
    throttle_classes = [AnonRateThrottle]
    serializer_class = PasswordForgetSerializer

    def post(self, request: Request):
        """Send a forgot password reset email code."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj: UserEmail = serializer.save()

        # check bad request rules
        check_security_rules(saas_settings.RESET_PASSWORD_SECURITY_RULES, request)

        code = serializer.save_password_code(obj)
        name = obj.user.get_full_name() or obj.user.get_username()
        recipients = [formataddr((name, obj.email))]
        self.send_email(recipients, code=code, user=obj.user)
        return Response(status=204)


class PasswordResetEndpoint(Endpoint):
    authentication_classes = []
    permission_classes = []
    throttle_classes = [AnonRateThrottle]
    serializer_class = PasswordResetSerializer

    def post(self, request: Request):
        """Reset password of a user with the given code."""
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response({'next': settings.LOGIN_URL})



from rest_framework import serializers
from ..models import get_tenant_model
from ..models import Permission, Group
from ..signals import (
    before_create_tenant,
    before_update_tenant,
)


class PermissionSerializer(serializers.ModelSerializer):
    class Meta:
        model = Permission
        fields = ['name', 'description', 'internal']


class TenantSerializer(serializers.ModelSerializer):
    class Meta:
        model = get_tenant_model()
        exclude = ['owner']

    def create(self, validated_data):
        before_create_tenant.send(self.__class__, data=validated_data, **self.context)
        return super().create(validated_data)

    def update(self, instance, validated_data):
        before_update_tenant.send(self.__class__, tenant=instance, data=validated_data, **self.context)
        return super().update(instance, validated_data)


class GroupSerializer(serializers.ModelSerializer):
    class Meta:
        model = Group
        fields = ['id', 'name', 'managed']


class GroupPermissionSerializer(serializers.ModelSerializer):
    permissions = PermissionSerializer(many=True)

    class Meta:
        model = Group
        fields = ['id', 'name', 'permissions', 'managed']



from django.http import HttpResponse, Http404
from django.contrib.auth.models import User
from ..models import Tenant, Member
from ..mail import render_mail_messages
from ..settings import saas_settings


def view_signup_code(request, suffix: str):
    context = {
        'code': 'ABCD',
        'site': saas_settings.SITE,
    }
    return _render_response('signup_code', context, suffix)


def view_reset_password(request, suffix: str):
    context = {
        'code': 'ABCD',
        'site': saas_settings.SITE,
        'user': User(username='alice', first_name='Alice'),
    }
    return _render_response('reset_password', context, suffix)


def view_invite_member(request, suffix: str):
    tenant = Tenant(slug='acme')
    context = dict(
        site=saas_settings.SITE,
        inviter=User(username='alice', first_name='Alice'),
        member=Member(name='Bob', invite_email='bob@example.com', tenant=tenant),
        tenant=tenant,
        invite_link='#',
    )
    return _render_response('invite_member', context, suffix)


def _render_response(template, context, suffix: str):
    text_message, html_message = render_mail_messages(template, context)
    if suffix == 'txt':
        return HttpResponse(text_message, content_type='text/plain')
    elif suffix == 'html':
        return HttpResponse(html_message, content_type='text/html')
    else:
        raise Http404



from django.utils.translation import gettext as _
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from ..drf.serializers import ModelSerializer
from ..models import Member, UserEmail
from .tenant import (
    PermissionSerializer,
    GroupPermissionSerializer,
    TenantSerializer,
)
from .user import UserSerializer


class MemberSerializer(ModelSerializer):
    user = UserSerializer(required=False, read_only=True)
    inviter = UserSerializer(required=False, read_only=True)
    groups = GroupPermissionSerializer(required=False, many=True, read_only=True)
    permissions = PermissionSerializer(required=False, many=True, read_only=True)

    class Meta:
        model = Member
        exclude = ['tenant']
        include_only_fields = ['user', 'groups', 'permissions']


class MemberInviteSerializer(ModelSerializer):
    invite_email = serializers.EmailField(required=True)

    class Meta:
        model = Member
        fields = ['name', 'invite_email', 'is_owner']

    def validate_invite_email(self, email: str):
        view = self.context['view']
        tenant_id = view.get_tenant_id()
        if Member.objects.filter(tenant_id=tenant_id, invite_email=email).count():
            raise ValidationError(_('This email has already been invited.'))
        return email

    def create(self, validated_data):
        email = validated_data['invite_email']
        try:
            user_email = UserEmail.objects.get_by_email(email)
            validated_data['user_id'] = user_email.user_id
            validated_data['status'] = Member.InviteStatus.WAITING
        except UserEmail.DoesNotExist:
            pass
        request = self.context['request']
        validated_data['inviter'] = request.user
        return super().create(validated_data)


class MemberDetailSerializer(ModelSerializer):
    groups = GroupPermissionSerializer(many=True, read_only=True)
    permissions = PermissionSerializer(many=True, read_only=True)

    class Meta:
        model = Member
        exclude = ['tenant', 'user']


class UserTenantsSerializer(ModelSerializer):
    tenant = TenantSerializer()
    groups = GroupPermissionSerializer(many=True, read_only=True)
    permissions = PermissionSerializer(many=True, read_only=True)

    class Meta:
        model = Member
        exclude = ['user', 'inviter', 'invite_email']



from rest_framework import serializers
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import OpenApiParameter, OpenApiExample, extend_schema
from drf_spectacular.extensions import OpenApiViewExtension


class AuthResponseSerializer(serializers.Serializer):
    next = serializers.CharField()


class PasswordLogInEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.auth.PasswordLogInEndpoint'

    def view_replacement(self):
        class FixedPasswordLogInEndpoint(self.target_class):
            @extend_schema(summary='Log In', responses={200: AuthResponseSerializer})
            def post(self, *args, **kwargs):
                pass

        return FixedPasswordLogInEndpoint


class LogoutEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.auth.LogoutEndpoint'

    def view_replacement(self):
        class FixedLogoutEndpoint(self.target_class):
            @extend_schema(summary='Log Out', request=None, responses={200: AuthResponseSerializer})
            def post(self, *args, **kwargs):
                pass

        return FixedLogoutEndpoint


class SignupConfirmEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.auth.SignupConfirmEndpoint'

    def view_replacement(self):
        class FixedSignupConfirmEndpoint(self.target_class):
            @extend_schema(summary='Sign Up', responses={200: AuthResponseSerializer})
            def post(self, *args, **kwargs):
                pass

        return FixedSignupConfirmEndpoint


class SignupRequestEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.auth.SignupRequestEndpoint'

    def view_replacement(self):
        class FixedSignupRequestEndpoint(self.target_class):
            @extend_schema(summary='Request to Sign-up', responses={204: None})
            def post(self, *args, **kwargs):
                pass

        return FixedSignupRequestEndpoint


class PasswordResetEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.password.PasswordResetEndpoint'

    def view_replacement(self):
        class FixedPasswordResetEndpoint(self.target_class):
            @extend_schema(summary='Password Reset', responses={200: AuthResponseSerializer})
            def post(self, *args, **kwargs):
                pass

        return FixedPasswordResetEndpoint


class PasswordForgotEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.password.PasswordForgotEndpoint'

    def view_replacement(self):
        class FixedPasswordForgotEndpoint(self.target_class):
            @extend_schema(summary='Password Forgot', responses={204: None})
            def post(self, *args, **kwargs):
                pass

        return FixedPasswordForgotEndpoint


id_parameter = OpenApiParameter(
    'id',
    type=str,
    location=OpenApiParameter.PATH,
    required=True,
)

member_id_parameter = OpenApiParameter(
    'member_id',
    type=str,
    location=OpenApiParameter.PATH,
    required=True,
)


class MemberGroupsEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.members.MemberGroupsEndpoint'

    def view_replacement(self):
        class Fixed(self.target_class):
            @extend_schema(
                operation_id='members_groups_list',
                summary='List Member Groups',
                parameters=[member_id_parameter],
            )
            def get(self, *args, **kwargs):
                pass

            @extend_schema(
                operation_id='members_groups_reset',
                summary='Reset Member Groups',
                parameters=[member_id_parameter],
                request=OpenApiTypes.UUID,
                responses={204: None},
                examples=[
                    OpenApiExample(
                        'Example 1',
                        value=['497f6eca-6276-4993-bfeb-53cbbbba6f08'],
                        request_only=True,
                    ),
                ],
            )
            def post(self, *args, **kwargs):
                pass

        return Fixed


class MemberGroupItemEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.members.MemberGroupItemEndpoint'

    def view_replacement(self):
        class Fixed(self.target_class):
            @extend_schema(
                operation_id='members_groups_remove',
                summary='Remove Member Group',
                parameters=[member_id_parameter, id_parameter],
            )
            def delete(self, *args, **kwargs):
                pass

        return Fixed


class MemberPermissionsEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.members.MemberPermissionsEndpoint'

    def view_replacement(self):
        class Fixed(self.target_class):
            @extend_schema(
                operation_id='members_permissions_list',
                summary='List Member Permissions',
                parameters=[member_id_parameter],
            )
            def get(self, *args, **kwargs):
                pass

            @extend_schema(
                operation_id='members_permissions_reset',
                summary='Reset Member Permissions',
                parameters=[member_id_parameter],
                request=OpenApiTypes.UUID,
                responses={204: None},
                examples=[
                    OpenApiExample(
                        'Example 1',
                        value=['tenant.read', 'tenant.admin'],
                        request_only=True,
                    ),
                ],
            )
            def post(self, *args, **kwargs):
                pass

        return Fixed


class MemberPermissionItemEndpoint(OpenApiViewExtension):
    target_class = 'saas_base.endpoints.members.MemberPermissionItemEndpoint'

    def view_replacement(self):
        class Fixed(self.target_class):
            @extend_schema(
                operation_id='members_permissions_remove',
                summary='Remove Member Permission',
                parameters=[member_id_parameter, id_parameter],
            )
            def delete(self, *args, **kwargs):
                pass

        return Fixed



from django.urls import path
from ..endpoints.members import (
    MemberListEndpoint,
    MemberItemEndpoint,
    MemberGroupsEndpoint,
    MemberGroupItemEndpoint,
    MemberPermissionsEndpoint,
    MemberPermissionItemEndpoint,
)

urlpatterns = [
    path('', MemberListEndpoint.as_view()),
    path('<pk>/', MemberItemEndpoint.as_view()),
    path('<member_id>/groups/', MemberGroupsEndpoint.as_view()),
    path('<member_id>/permissions/', MemberPermissionsEndpoint.as_view()),
    path('<member_id>/groups/<pk>/', MemberGroupItemEndpoint.as_view()),
    path('<member_id>/permissions/<pk>/', MemberPermissionItemEndpoint.as_view()),
]



import typing as t
from .base import BaseMailProvider, render_mail_messages
from .django import DjangoMailProvider
from ..signals import mail_queued
from ..settings import saas_settings


def get_mail_provider(name: str = 'default') -> t.Optional[BaseMailProvider]:
    if name in saas_settings.MAIL_PROVIDERS:
        return saas_settings.MAIL_PROVIDERS[name]


class SendEmailMixin:
    email_provider: str = 'default'
    email_subject: str
    email_template_id: str

    def get_email_subject(self):
        return self.email_subject

    def send_email(self, recipients, **context):
        context.setdefault('site', saas_settings.SITE)
        if saas_settings.MAIL_IMMEDIATE_SEND:
            provider = get_mail_provider(self.email_provider)
            provider.send_context_mail(
                subject=self.get_email_subject(),
                template_id=self.email_template_id,
                recipients=recipients,
                context=context,
            )
        else:
            mail_queued.send(
                sender=self.__class__,
                subject=self.get_email_subject(),
                template_id=self.email_template_id,
                recipients=recipients,
                context=context,
            )


__all__ = [
    'get_mail_provider',
    'render_mail_messages',
    'BaseMailProvider',
    'DjangoMailProvider',
    'SendEmailMixin',
]



from django.template import loader
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _
from django.conf import settings
from rest_framework.filters import BaseFilterBackend
from .errors import BadRequest
from ..settings import saas_settings


class TenantIdFilter(BaseFilterBackend):
    tenant_id_field = 'tenant_id'

    def filter_queryset(self, request, queryset, view):
        query_field = getattr(view, 'tenant_id_field', self.tenant_id_field)

        tenant_id = getattr(request, 'tenant_id', None)
        if not tenant_id:
            raise BadRequest('Missing Tenant ID')

        kwargs = {query_field: tenant_id}
        return queryset.filter(**kwargs)

    def get_schema_operation_parameters(self, view):
        in_header = 'saas_base.middleware.HeaderTenantIdMiddleware' in settings.MIDDLEWARE
        in_path = 'saas_base.middleware.PathTenantIdMiddleware' in settings.MIDDLEWARE

        parameters = []
        if in_header:
            parameters.append(
                {
                    'name': saas_settings.TENANT_ID_HEADER,
                    'required': True,
                    'in': 'header',
                    'schema': {
                        'type': 'string',
                    },
                }
            )

        if in_path:
            parameters.append(
                {
                    'name': 'tenant_id',
                    'required': True,
                    'in': 'path',
                    'schema': {
                        'type': 'string',
                    },
                }
            )

        if len(parameters) == 2:
            parameters[0]['required'] = False
            parameters[0]['description'] = 'Tenant ID (required if not provided in path).'
            parameters[1]['required'] = False
            parameters[1]['description'] = 'Tenant ID (required if not provided in header).'

        return parameters


class IncludeFilter(BaseFilterBackend):
    template = 'saas/include_filter.html'
    include_title = _('Include')
    include_description = _('A include term.')

    def get_select_related_fields(self, view, request):
        return getattr(view, 'include_select_related_fields', [])

    def get_prefetch_related_fields(self, view, request):
        return getattr(view, 'include_prefetch_related_fields', [])

    def get_include_terms(self, request):
        params = request.query_params.get('include', '')
        params = params.replace('\x00', '')  # strip null characters
        params = params.replace(',', ' ')
        return params.split()

    def filter_queryset(self, request, queryset, view):
        select_related_fields = self.get_select_related_fields(view, request)
        prefetch_related_fields = self.get_prefetch_related_fields(view, request)
        if not select_related_fields and not prefetch_related_fields:
            return queryset

        include_terms = self.get_include_terms(request)
        if not include_terms:
            return queryset

        if include_terms == ['all']:
            include_terms = select_related_fields + prefetch_related_fields

        for field in include_terms:
            if field in select_related_fields:
                queryset = queryset.select_related(field)
            elif field in prefetch_related_fields:
                queryset = queryset.prefetch_related(field)

        request.include_terms = include_terms
        return queryset

    def to_html(self, request, queryset, view):
        context = {'term': request.query_params.get('include', '')}
        template = loader.get_template(self.template)
        return template.render(context)

    def get_schema_operation_parameters(self, view):
        return [
            {
                'name': 'include',
                'required': False,
                'in': 'query',
                'description': force_str(self.include_description),
                'schema': {
                    'type': 'string',
                },
            },
        ]



from django.utils.functional import SimpleLazyObject
from django.conf import settings
from .settings import saas_settings
from .models import get_tenant_model

__all__ = [
    'TenantMiddleware',
    'ConfiguredTenantIdMiddleware',
    'HeaderTenantIdMiddleware',
    'PathTenantIdMiddleware',
    'SessionTenantIdMiddleware',
]
TenantModel = get_tenant_model()


class TenantMiddleware:
    def __init__(self, get_response=None):
        self.get_response = get_response

    @staticmethod
    def get_tenant(request):
        if not hasattr(request, '_cached_tenant'):
            TenantModel = get_tenant_model()
            tenant_id = getattr(request, 'tenant_id', None)
            if not tenant_id:
                return None
            try:
                tenant = TenantModel.objects.get_from_cache_by_pk(tenant_id)
                request._cached_tenant = tenant
            except TenantModel.DoesNotExist:
                return None
        return getattr(request, '_cached_tenant', None)

    def __call__(self, request):
        request.tenant = SimpleLazyObject(lambda: self.get_tenant(request))
        response = self.get_response(request)
        return response


class ConfiguredTenantIdMiddleware:
    SETTING_KEY = 'SAAS_TENANT_ID'

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        request.tenant_id = getattr(settings, self.SETTING_KEY, None)
        return self.get_response(request)


class HeaderTenantIdMiddleware:
    HTTP_HEADER = saas_settings.TENANT_ID_HEADER

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        if getattr(request, 'tenant_id', None):
            return self.get_response(request)

        tenant_id = request.headers.get(self.HTTP_HEADER)
        request.tenant_id = TenantModel._meta.pk.to_python(tenant_id)
        return self.get_response(request)


class PathTenantIdMiddleware:
    FIELD_KEY = 'tenant_id'

    def __init__(self, get_response=None):
        self.get_response = get_response

    def process_view(self, request, view_func, view_args, view_kwargs):
        if getattr(request, 'tenant_id', None):
            return
        tenant_id = view_kwargs.get(self.FIELD_KEY)
        request.tenant_id = TenantModel._meta.pk.to_python(tenant_id)

    def __call__(self, request):
        return self.get_response(request)


class SessionTenantIdMiddleware:
    FIELD_KEY = 'tenant_id'

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        if getattr(request, 'tenant_id', None):
            return self.get_response(request)

        tenant_id = request.session.get(self.FIELD_KEY)
        request.tenant_id = TenantModel._meta.pk.to_python(tenant_id)
        return self.get_response(request)



from saas_base.test import SaasTestCase


class TestUserPasswordAPI(SaasTestCase):
    user_id = SaasTestCase.OWNER_USER_ID

    def send_request(self, data):
        self.force_login()
        url = '/m/user/password/'
        return self.client.post(url, data=data, format='json')

    def test_update_password(self):
        data = {'password': 'abc12.3D', 'confirm_password': 'abc12.3D'}
        resp = self.send_request(data)
        self.assertEqual(resp.status_code, 204)

    def test_not_match_password(self):
        data = {'password': 'abc12.3D', 'confirm_password': 'abc12.3C'}
        resp = self.send_request(data)
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.json()['password'], ['Password does not match.'])

    def test_too_simple_password(self):
        data = {'password': 'foo', 'confirm_password': 'foo'}
        resp = self.send_request(data)
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(
            resp.json()['password'], ['This password is too short. It must contain at least 8 characters.']
        )



import uuid
from django.db import models
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from ..db import CachedManager


class EmailManager(CachedManager):
    natural_key = ['email']

    def get_by_email(self, email: str) -> 'UserEmail':
        return self.get_from_cache_by_natural_key(email)


class UserEmail(models.Model):
    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name='emails')
    email = models.EmailField(unique=True)
    verified = models.BooleanField(default=False)
    primary = models.BooleanField(default=False, db_index=True)
    created_at = models.DateTimeField(default=timezone.now, db_index=True)

    objects = EmailManager()

    class Meta:
        verbose_name = _('email')
        verbose_name_plural = _('emails')
        ordering = ['created_at']
        db_table = 'saas_auth_email'

    def __str__(self):
        return self.email

    def natural_key(self):
        return (self.email,)



from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.mixins import ListModelMixin
from ..drf.views import AuthenticatedEndpoint
from ..models import UserEmail, Member
from ..serializers.user import (
    UserSerializer,
    UserEmailSerializer,
    UserPasswordSerializer,
)
from ..serializers.member import UserTenantsSerializer


class UserEndpoint(AuthenticatedEndpoint):
    resource_scopes = ['user', 'user:profile']
    serializer_class = UserSerializer

    def get(self, request: Request):
        """Retrieve current user information."""
        serializer: UserSerializer = self.get_serializer(request.user)
        return Response(serializer.data)


class UserPasswordEndpoint(AuthenticatedEndpoint):
    resource_scopes = ['user:password']
    serializer_class = UserPasswordSerializer

    def post(self, request: Request, *args, **kwargs):
        """Update current user's password"""
        serializer: UserPasswordSerializer = self.get_serializer(request.user, data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        return Response(status=204)


class UserEmailListEndpoint(ListModelMixin, AuthenticatedEndpoint):
    resource_scopes = ['user:email']
    pagination_class = None
    serializer_class = UserEmailSerializer

    def get_queryset(self):
        return UserEmail.objects.filter(user=self.request.user).all()

    def get(self, request: Request, *args, **kwargs):
        """List all the current user's emails."""
        return self.list(request, *args, **kwargs)

    def post(self, request: Request, *args, **kwargs):
        # add email
        pass


class UserTenantsEndpoint(ListModelMixin, AuthenticatedEndpoint):
    serializer_class = UserTenantsSerializer
    queryset = Member.objects.select_related('tenant').all()

    def filter_queryset(self, queryset):
        status = self.request.query_params.get('status')
        queryset = queryset.prefetch_related('groups', 'permissions', 'groups__permissions')
        queryset = queryset.filter(user=self.request.user)

        if status == 'all':
            return queryset.all()
        if status == 'waiting':
            queryset = queryset.filter(status=Member.InviteStatus.WAITING)
        else:
            queryset = queryset.filter(status=Member.InviteStatus.ACTIVE)
        return queryset.all()

    def get(self, request: Request, *args, **kwargs):
        """List all the current user's tenants."""
        return self.list(request, *args, **kwargs)



import re
import typing as t
from collections import OrderedDict
from rest_framework.fields import ChoiceField as _ChoiceField
from rest_framework.serializers import ModelSerializer as _ModelSerializer


class ChoiceField(_ChoiceField):
    """Rewrite this fields to support (int, str) choices."""

    def __init__(self, choices, **kwargs):
        super().__init__(choices, **kwargs)

        self.int_str_choices: t.Dict[int, str] = {}
        self.str_int_choices: t.Dict[str, int] = {}
        if choices:
            self._set_int_str_choices(choices)

    def _set_int_str_choices(self, choices: t.List[t.Tuple[int, str]]):
        int_str_choices: t.Dict[int, str] = {}
        str_int_choices: t.Dict[str, int] = {}
        for choice in choices:
            if len(choice) == 2 and isinstance(choice[0], int) and _is_lower_string(choice[1]):
                int_str_choices[choice[0]] = choice[1]
                str_int_choices[choice[1]] = choice[0]
            else:
                return
        self.int_str_choices = int_str_choices
        self.str_int_choices = str_int_choices

    def to_internal_value(self, data):
        if data == '' and self.allow_blank:
            return ''

        if self.str_int_choices:
            try:
                return self.str_int_choices[str(data)]
            except KeyError:
                self.fail('invalid_choice', input=data)

        try:
            return self.choice_strings_to_values[str(data)]
        except KeyError:
            self.fail('invalid_choice', input=data)

    def to_representation(self, value):
        if value in ('', None):
            return value

        if self.int_str_choices:
            return self.int_str_choices[value]

        return self.choice_strings_to_values.get(str(value), value)


class ModelSerializer(_ModelSerializer):
    serializer_choice_field = ChoiceField

    @property
    def _readable_fields(self):
        include_only_fields = getattr(self.Meta, 'include_only_fields', [])
        if not include_only_fields:
            for field in self.fields.values():
                if not field.write_only:
                    yield field
        else:
            include_terms = self.context.get('include_fields', [])
            if not include_terms:
                request = self.context.get('request')
                if request:
                    include_terms = getattr(request, 'include_terms', [])

            for field in self.fields.values():
                if field.field_name in include_only_fields and field.field_name not in include_terms:
                    continue

                if not field.write_only:
                    yield field

    def get_fields(self):
        fields = super().get_fields()
        flatten_fields = getattr(self.Meta, 'flatten_fields', None)
        if flatten_fields:
            return _make_flatten_fields(fields, flatten_fields)
        return fields


def _make_flatten_fields(form: OrderedDict, flatten: t.List[str]):
    for source in flatten:
        serializer: t.Any = form.pop(source, None)
        if serializer:
            fields = serializer.get_fields()
            for name in fields:
                field = fields[name]
                field.source = f'{source}.{name}'
                if name in form:
                    form[f'{source}_{name}'] = field
                else:
                    form[name] = field
    return form


def _is_lower_string(s: str):
    return isinstance(s, str) and bool(re.match(r'^[0-9a-z-_]+$', s))



import typing as t
import logging
from rest_framework.request import Request
from .rules import Rule, BlockedEmailDomains, TooManyDots, Turnstile
from ..drf.errors import BadRequest

__all__ = [
    'Rule',
    'BlockedEmailDomains',
    'TooManyDots',
    'Turnstile',
    'check_security_rules',
]

logger = logging.getLogger(__name__)


def check_security_rules(security_rules: t.List[Rule], request: Request):
    for rule in security_rules:
        if rule.bad_request(request):
            rule_name = rule.__class__.__name__
            logger.warning(f'Bad request: [{rule_name}].')
            raise BadRequest()



import typing as t
from django.core.mail import EmailMultiAlternatives, get_connection
from .base import BaseMailProvider


class SMTPProvider(BaseMailProvider):
    name = 'smtp'

    def get_connection(self, fail_silently: bool = False):
        return get_connection(fail_silently=fail_silently, **self.options)

    def send_mail(
        self,
        subject: str,
        recipients: t.List[str],
        text_message: str,
        html_message: t.Optional[str] = None,
        from_email: t.Optional[str] = None,
        headers: t.Optional[t.Dict[str, str]] = None,
        reply_to: t.Optional[str] = None,
        fail_silently: bool = False,
    ):
        if from_email is None:
            from_email = self.default_from_email
        connection = self.get_connection(fail_silently=fail_silently)
        mail = EmailMultiAlternatives(
            subject,
            body=text_message,
            from_email=from_email,
            to=recipients,
            connection=connection,
            headers=headers,
            reply_to=reply_to,
        )
        if html_message:
            mail.attach_alternative(html_message, 'text/html')
        return mail.send()



from saas_base.models import Member, Permission, Group
from saas_base.models import get_tenant_model
from saas_base.test import SaasTestCase


class TestTenantsAPI(SaasTestCase):
    user_id = SaasTestCase.EMPTY_USER_ID

    def test_list_tenants(self):
        self.force_login()
        user = self.get_user()

        tenants = create_demo_tenants(count=2)
        for tenant in tenants:
            Member.objects.create(
                user=user,
                name=user.get_full_name(),
                tenant=tenant,
                status=Member.InviteStatus.WAITING,
            )

        url = '/m/user/tenants/?status=all'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()['count'], 2)

        url = '/m/user/tenants/?status=waiting'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()['count'], 2)

        url = '/m/user/tenants/'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()['count'], 0)

    def test_list_tenant_flat_permissions(self):
        self.force_login()
        user = self.get_user()
        tenants = create_demo_tenants(count=2)

        members = []
        for tenant in tenants:
            member = Member.objects.create(
                user=user,
                tenant=tenant,
                name=user.get_full_name(),
                status=Member.InviteStatus.ACTIVE,
            )
            members.append(member)
            member.permissions.add(Permission.objects.get_by_name('tenant.read'))

        group = Group.objects.create(tenant=tenants[0], name='Admin')
        group.permissions.add(Permission.objects.get_by_name('tenant.admin'))
        members[0].groups.add(group)

        group = Group.objects.create(tenant=tenants[1], name='Guest')
        members[1].groups.add(group)

        url = '/m/user/tenants/'
        resp = self.client.get(url)
        self.assertEqual(resp.status_code, 200)
        data = resp.json()
        self.assertEqual(data['count'], 2)


def create_demo_tenants(prefix: str = 'demo', region: str = 'us', count: int = 10):
    tenants = []
    for i in range(count):
        tenant = get_tenant_model().objects.create(slug=f'{prefix}-0{i}', region=region)
        tenants.append(tenant)
    return tenants



import requests
from abc import ABCMeta, abstractmethod
from rest_framework.request import Request
from .ip import get_client_ip


class Rule(metaclass=ABCMeta):
    DEFAULT_RESPONSE_FIELD: str = 'email'

    def __init__(self, **options):
        self.options = options

    def get_response_field_value(self, request: Request):
        response_field = self.options.get(
            'response_field',
            self.DEFAULT_RESPONSE_FIELD,
        )
        return request.data.get(response_field)

    @abstractmethod
    def bad_request(self, request: Request):
        pass


class BlockedEmailDomains(Rule):
    def bad_request(self, request: Request):
        blocked_list = self.options.get('domains')
        email = self.get_response_field_value(request)
        return blocked_list and email.endswith(tuple([f'@{s}' for s in blocked_list]))


class TooManyDots(Rule):
    MAX_DOT_COUNT = 4

    def bad_request(self, request: Request):
        max_dot_count = self.options.get('count', self.MAX_DOT_COUNT)
        email = self.get_response_field_value(request)
        name = email.split('@')[0]
        return name.count('.') > max_dot_count


class Turnstile(Rule):
    API_URL = 'https://challenges.cloudflare.com/turnstile/v0/siteverify'
    DEFAULT_RESPONSE_FIELD = 'cf-turnstile-response'

    def bad_request(self, request: Request):
        token = self.get_response_field_value(request)
        if not token:
            return True

        secret = self.options.get('secret')
        ip = get_client_ip(request)
        data = {'secret': secret, 'remoteip': ip, 'response': token}
        resp = requests.post(self.API_URL, data=data, timeout=5)
        resp.raise_for_status()
        data = resp.json()
        return not data.get('success')



from saas_base.models import Group, Permission
from saas_base.test import SaasTestCase


class TestAccountScopes(SaasTestCase):
    user_id = SaasTestCase.GUEST_USER_ID

    @classmethod
    def setUpTestData(cls):
        Permission.objects.initialize_names(['account'])

    def test_none_user_permissions(self):
        member = self.get_user_member()
        self.assertEqual(member.user_permissions, [])

    def test_exist_user_permissions(self):
        self.add_user_perms('account.read', 'account.write', 'account.admin')
        member = self.get_user_member()
        scopes = set(member.user_permissions)
        self.assertEqual(scopes, {'account.read', 'account.write', 'account.admin'})

    def test_none_group_permissions(self):
        member = self.get_user_member()
        self.assertEqual(member.group_permissions, [])

    def test_exist_group_permissions(self):
        group = Group.objects.create(tenant=self.tenant, name='Group')
        perms = [item for item in Permission.objects.all() if item.name.startswith('account.')]
        group.permissions.add(*perms)

        member = self.get_user_member()
        member.groups.add(group)
        scopes = set(member.group_permissions)
        self.assertEqual(scopes, {'account.read', 'account.write', 'account.admin'})
